//
// Copyright (C) 2023 Intel Corporation.
// SPDX-License-Identifier: Apache 2.0
//

#pragma once

#include <openvino/runtime/iinfer_request.hpp>

#include "variable_state.hpp"
#include "vpux_compiler.hpp"

namespace vpux {

/**
 * @brief Acts as an interface for the inference request structures implemented by all backends.
 * @details The operations common for all backends can be found implemented here, these include tensors' extraction,
 * state variables handling and and a few helper functions.
 * @note Regarding the code design, the "ov::ISyncInferRequest" would normally be the better option for
 * inheritance. However, the interface exposed by that class forces some additional latency in several unfavorable
 * scenarios, thus a reimplementation was required.
 */
class SyncInferRequest : public ov::IInferRequest {
public:
    explicit SyncInferRequest(const std::shared_ptr<const ov::ICompiledModel> compiled_model,
                              const std::shared_ptr<const NetworkDescription> networkDescription);

    /**
     * @brief Gets an input/output tensor for inference.
     * @note If the tensor with the specified @p port is not found, an exception is thrown.
     * @param port Port of the tensor to get.
     * @return Tensor for the port @p port.
     */
    ov::SoPtr<ov::ITensor> get_tensor(const ov::Output<const ov::Node>& port) const override;

    /**
     * @brief Sets an input/output tensor to infer.
     * @param port Port of the input or output tensor.
     * @param tensor Reference to a tensor. The element_type and shape of a tensor must match
     * the model's input/output element_type and size.
     */
    void set_tensor(const ov::Output<const ov::Node>& port, const ov::SoPtr<ov::ITensor>& tensor) override;

    /**
     * @brief Currently there is no support implemented for batches of tensors, thus this call is a simple redirection
     * to the "get_tensor" one.
     */
    std::vector<ov::SoPtr<ov::ITensor>> get_tensors(const ov::Output<const ov::Node>& port) const override;

    /**
     * @brief Currently there is no support implemented for batches of tensors, thus this call is a simple redirection
     * to the "set_tensor" one.
     */
    void set_tensors(const ov::Output<const ov::Node>& port,
                     const std::vector<ov::SoPtr<ov::ITensor>>& tensors) override;

    /**
     * @brief Gets inputs for infer request
     *
     * @return vector of input ports
     */
    const std::vector<ov::Output<const ov::Node>>& get_inputs() const override;

    /**
     * @brief Gets outputs for infer request
     *
     * @return vector of output ports
     */
    const std::vector<ov::Output<const ov::Node>>& get_outputs() const override;

    /**
     * @brief Gets pointer to compiled model (usually synchronous request holds the compiled model)
     *
     * @return Pointer to the compiled model
     */
    const std::shared_ptr<const ov::ICompiledModel>& get_compiled_model() const override;

    /**
     * @brief Used for executing the inference.
     */
    virtual void infer_async() = 0;

    /**
     * @brief Used for retrieving the prediction's result.
     */
    virtual void get_result() = 0;

    std::vector<ov::SoPtr<ov::IVariableState>> query_state() const override;

    /**
     * @brief Initializes the tensor values corresponding to the state variables.
     * @details The inital values are usually all 0s.
     */
    void initialize_states();

    /**
     * @return The state tensors accessible by their names.
     */
    std::unordered_map<std::string, std::shared_ptr<VariableState>>& get_variable_states() {
        return _variableStates;
    }

    /**
     * @return The names used by the inputs in the order registered inside the model.
     */
    std::vector<std::string> get_input_names() {
        return _inputNames;
    }

    /**
     * @return The names used by the outputs in the order registered inside the model.
     */
    std::vector<std::string> get_output_names() {
        return _outputNames;
    }

    /**
     * @return The names used by the state variables in the order registered inside the model.
     */
    std::vector<std::string> get_state_names() {
        return _stateNames;
    }

    /**
     * @return A map holding references towards all tensors used by the current inference request object.
     */
    std::unordered_map<std::string, std::shared_ptr<ov::ITensor>>& get_all_tensors() {
        return _allTensors;
    }

protected:
    /**
     * @brief Basic checks for input/output tensor
     *
     * @param port Input/Output port
     * @param tensor Input/Output tensor
     */
    void check_tensor(const ov::Output<const ov::Node>& port, const ov::SoPtr<ov::ITensor>& tensor) const;

    /**
     * @brief Check that all tensors are valid. Throws an exception if it's not.
     */
    void check_tensors() const override;

    /**
     * @brief Checks if the provided precision value is supported by the current backend, should throw an error
     * otherwise.
     * @param precision The precision value to be checked.
     */
    virtual void check_network_precision(const ov::element::Type_t precision) = 0;

    /**
     * @brief Allocates a tensor on host and stores the reference inside the "_allTensors" attribute. If a buffer
     * address is provided, then the tensor is built upon it and no additional data buffer is allocated.
     * @param tensorName The name by which the tensor shall be identified
     * @param descriptor Tensor's metadata
     * @param dataBuffer If provided, the tensor shall point towards this data buffer instead of allocating another one.
     * @param isState If true, the tensor shall also be stored inside the state variables map. In this case, adding the
     * tensor to this structure would be required in order to correctly answer the state queries.
     */
    void allocate_tensor(std::string tensorName, const IONodeDescriptor& descriptor, void* dataBuffer,
                         const bool isState);

    // Mutable to return reference to ov::Tensor
    mutable std::unordered_map<std::string, std::shared_ptr<ov::ITensor>> _allTensors;
    std::unordered_map<std::string, std::shared_ptr<VariableState>> _variableStates;

    std::shared_ptr<const ov::ICompiledModel> _compiledModel;

    std::vector<std::string> _inputNames;
    std::vector<std::string> _outputNames;
    std::vector<std::string> _stateNames;

    // Stored in order to avoid additional processing when launching inferences
    std::vector<std::string> _inputAndStateInputNames;
    std::vector<std::string> _outputAndStateOutputNames;

    IONodeDescriptorMap _parameterDescriptors;
    IONodeDescriptorMap _resultDescriptors;
    IONodeDescriptorMap _stateDescriptors;

    std::unordered_map<std::string, std::string> _nodeNameToLegacyName;
};

}  // namespace vpux
