//
// Copyright (C) 2022 Intel Corporation.
// SPDX-License-Identifier: Apache 2.0
//

#include "vpux/compiler/dialect/IE/utils/resources.hpp"
#include "vpux/compiler/dialect/VPUIP/passes.hpp"

#include "vpux/compiler/dialect/VPUIP/convert_to_dma_utils.hpp"
#include "vpux/compiler/dialect/VPURT/ops.hpp"
#include "vpux/compiler/dialect/VPURT/task.hpp"
#include "vpux/compiler/utils/attributes.hpp"
#include "vpux/compiler/utils/rewriter.hpp"

#include <mlir/Transforms/GreedyPatternRewriteDriver.h>

using namespace vpux;

namespace {

//
// UpsamplingDMARewriter
//

class UpsamplingDMARewriter final : public mlir::OpRewritePattern<VPUIP::UpsamplingDMAOp> {
public:
    UpsamplingDMARewriter(mlir::MLIRContext* ctx, int64_t dmaPortCount, Logger log)
            : mlir::OpRewritePattern<VPUIP::UpsamplingDMAOp>(ctx), _log(log), _dmaPortCount(dmaPortCount) {
        setDebugName("UpsamplingDMARewriter");

        _cmxNameAttr = mlir::FlatSymbolRefAttr::get(ctx, stringifyEnum(VPU::MemoryKind::CMX_NN));
    }

    mlir::LogicalResult matchAndRewrite(VPUIP::UpsamplingDMAOp UpsamplingDMAOp,
                                        mlir::PatternRewriter& rewriter) const final;

private:
    Logger _log;
    int64_t _dmaPortCount;
    mlir::FlatSymbolRefAttr _cmxNameAttr;
};

static VPUIP::DMADescriptorAttr generateUpsamplingDmaDescriptor(mlir::MLIRContext* ctx, vpux::ShapeRef inShape,
                                                                mlir::ArrayAttr factor, Byte inElemTypeSize,
                                                                int64_t expandChannel) {
    auto upsampleFactor = parseIntArrayAttr<int64_t>(factor);
    auto elemTypeSize = inElemTypeSize.count();

    const auto IC = inShape[Dims4D::Act::C];
    const auto H = inShape[Dims4D::Act::H];
    const auto W = inShape[Dims4D::Act::W];
    auto len = vpux::getIntAttr(ctx, W * IC * elemTypeSize);
    auto srcWidth = vpux::getIntAttr(ctx, IC * W * elemTypeSize);
    auto srcStride = vpux::getIntAttr(ctx, IC * W * elemTypeSize);
    auto srcPlaneStride = vpux::getIntAttr(ctx, W * IC * elemTypeSize);
    auto dstWidth = vpux::getIntAttr(ctx, IC * elemTypeSize);
    auto dstStride = vpux::getIntAttr(ctx, (IC + expandChannel) * elemTypeSize * upsampleFactor[Dims4D::Act::W.ind()]);
    auto dstPlaneStride =
            vpux::getIntAttr(ctx, W * (IC + expandChannel) * elemTypeSize * upsampleFactor[Dims4D::Act::H.ind()] *
                                          upsampleFactor[Dims4D::Act::W.ind()]);
    auto numPlanes = vpux::getIntAttr(ctx, H);
    return VPUIP::DMADescriptorAttr::get(ctx, numPlanes, len, srcWidth, srcStride, srcPlaneStride, dstWidth, dstStride,
                                         dstPlaneStride);
}

void shapeReorder(VPUIP::UpsamplingDMAOp upsamplingDMAOp, mlir::PatternRewriter& rewriter,
                  vpux::NDTypeInterface inType) {
    auto finalShape = Shape(inType.getShape().raw());
    finalShape[Dims4D::Act::H] = finalShape[Dims4D::Act::H] * finalShape[Dims4D::Act::C];
    finalShape[Dims4D::Act::C] = 1;

    auto cst = upsamplingDMAOp.getInput().getDefiningOp<Const::DeclareOp>();
    const auto origConstType = cst.getType().cast<vpux::NDTypeInterface>();
    const auto newConstType = origConstType.changeShape(finalShape);
    const auto newConstAttr = cst.getContentAttr().reshape(finalShape);
    auto newCst = rewriter.create<Const::DeclareOp>(upsamplingDMAOp.getLoc(), newConstType, newConstAttr);
    rewriter.replaceOp(upsamplingDMAOp.getInput().getDefiningOp(), newCst.getOutput());
}

mlir::Value getNewSrcBuffer(VPUIP::UpsamplingDMAOp upsamplingDMAOp, mlir::PatternRewriter& rewriter,
                            vpux::Shape inputShape, int64_t& currentOffset) {
    auto cst = upsamplingDMAOp.getInput().getDefiningOp<Const::DeclareOp>();
    SmallVector<int64_t> offsetVec(inputShape.size(), 0);
    offsetVec[Dims4D::Act::H.ind()] = currentOffset;
    currentOffset += inputShape[Dims4D::Act::H];
    return rewriter.create<VPUIP::SubViewOp>(upsamplingDMAOp->getLoc(), cst, offsetVec, inputShape.raw());
}

mlir::LogicalResult UpsamplingDMARewriter::matchAndRewrite(VPUIP::UpsamplingDMAOp upsamplingDMAOp,
                                                           mlir::PatternRewriter& rewriter) const {
    _log.trace("Process UpsamplingDMA op: {0}", upsamplingDMAOp);
    if (upsamplingDMAOp.getDmaDescriptor().has_value()) {
        return mlir::failure();
    }
    auto parentOp = upsamplingDMAOp.getInput().getDefiningOp();
    auto vpurtTask = upsamplingDMAOp->getParentOfType<VPURT::TaskOp>();
    VPUX_THROW_UNLESS(vpurtTask != nullptr, "Can't get VPURT task operation");
    rewriter.setInsertionPointAfter(vpurtTask);

    auto inType = upsamplingDMAOp.getInput().getType().cast<vpux::NDTypeInterface>();
    Byte elemTypeSize = inType.getElemTypeSize();

    bool inputIsCst = false;
    auto srcDeclBuff = upsamplingDMAOp.getInput().getDefiningOp<VPURT::DeclareBufferOp>();
    VPUX_THROW_UNLESS((mlir::isa<VPURT::DeclareBufferOp, Const::DeclareOp>(parentOp)),
                      "Can't get buffer for operand: {0}", upsamplingDMAOp.getInput());
    auto dstDeclBuff = upsamplingDMAOp.getOutputBuff().getDefiningOp<VPURT::DeclareBufferOp>();
    VPUX_THROW_UNLESS(dstDeclBuff != nullptr, "Can't get buffer for operand: {0}", upsamplingDMAOp.getOutputBuff());

    inputIsCst = mlir::isa<Const::DeclareOp>(parentOp);
    auto srcType = upsamplingDMAOp.getInput().getType().cast<vpux::NDTypeInterface>();
    auto dstType = dstDeclBuff.getType().cast<vpux::NDTypeInterface>();
    const auto inOrder = DimsOrder::fromValue(upsamplingDMAOp.getInput());
    auto subShape = Shape(inType.getShape().raw());
    int64_t totalNumPlane = inType.getShape()[Dims4D::Act::N] * inType.getShape()[Dims4D::Act::H];

    if (inOrder == DimsOrder::NCHW) {
        totalNumPlane *= inType.getShape()[Dims4D::Act::C];
        subShape[Dims4D::Act::C] = 1;
        if (inputIsCst) {
            // Strides matching with shape and order
            shapeReorder(upsamplingDMAOp, rewriter, inType);
        }
    }

    auto fullCopySize = static_cast<Byte>(getCompactSize(upsamplingDMAOp.getInput()));
    auto numberDMAsPlanesRestriction = divUp(totalNumPlane, VPUIP::DMA_MAX_NUMBER_PLANES);
    auto numberDMAsSizeRestriction = divUp(fullCopySize.count(), VPUIP::DMA_LIMIT.count());
    auto numberDMAs = std::max(numberDMAsPlanesRestriction, numberDMAsSizeRestriction);
    auto singlePlaneSize = fullCopySize.count() / totalNumPlane;
    auto maxNumberOfPlanesPerDMA = std::min(VPUIP::DMA_LIMIT.count() / singlePlaneSize, totalNumPlane);
    auto numberPlanesPerDMA = std::min(VPUIP::DMA_MAX_NUMBER_PLANES, maxNumberOfPlanesPerDMA);
    subShape[Dims4D::Act::N] = 1;
    subShape[Dims4D::Act::H] = numberPlanesPerDMA;

    SmallVector<Shape> subInputShapes(numberDMAs - 1, subShape);
    subShape[Dims4D::Act::H] = totalNumPlane - (numberPlanesPerDMA * (numberDMAs - 1));
    subInputShapes.push_back(subShape);
    auto dstOffset = dstDeclBuff.getByteOffset();

    auto context = upsamplingDMAOp.getContext();
    auto upsamplingFactor = parseIntArrayAttr<int64_t>(upsamplingDMAOp.getUpsamplingFactor());
    auto hasExpandAttr = upsamplingDMAOp.getExpand().has_value();
    SmallVector<int64_t, 4> expand;
    if (hasExpandAttr) {
        expand = mlir::extractFromIntegerArrayAttr<int64_t>(upsamplingDMAOp.getExpandAttr());
    }
    auto getOutShape = [&upsamplingFactor, &hasExpandAttr, &expand](ShapeRef inShape) {
        auto outShape = Shape(inShape.raw());
        for (size_t i = 0; i < outShape.size(); i++) {
            outShape[Dim(i)] *= upsamplingFactor[i];
            if (hasExpandAttr) {
                outShape[Dim(i)] += expand[i];
            }
        }
        return outShape;
    };
    int64_t dmaPort = 0;
    int64_t srcOffset = 0, currentOffset = 0;
    if (!inputIsCst) {
        srcOffset = srcDeclBuff.getByteOffset();
    }

    for (auto& inputShape : subInputShapes) {
        mlir::Value newSrcBuff;
        if (!inputIsCst) {
            auto newSrcMemRef =
                    vpux::getMemRefType(inputShape, srcType.getElementType(), inOrder, srcType.getMemSpace());
            newSrcBuff = VPUIP::createNewDeclareBuffer(rewriter, srcDeclBuff, srcDeclBuff, newSrcMemRef, srcOffset);
        } else {
            newSrcBuff = getNewSrcBuffer(upsamplingDMAOp, rewriter, inputShape, currentOffset);
        }

        auto outShape = getOutShape(inputShape);
        auto newDstMemRef = vpux::getMemRefType(outShape, dstType.getElementType(), inOrder, dstType.getMemSpace());
        auto newDstBuff = VPUIP::createNewDeclareBuffer(rewriter, dstDeclBuff, dstDeclBuff, newDstMemRef, dstOffset);
        auto descriptorAttr =
                hasExpandAttr
                        ? generateUpsamplingDmaDescriptor(context, inputShape, upsamplingDMAOp.getUpsamplingFactor(),
                                                          inType.getElemTypeSize(), expand[1])
                        : generateUpsamplingDmaDescriptor(context, inputShape, upsamplingDMAOp.getUpsamplingFactor(),
                                                          inType.getElemTypeSize(), 0);

        auto nextOffset = srcOffset + inputShape.totalSize() * elemTypeSize.count();
        const auto newLoc =
                appendLoc(upsamplingDMAOp->getLoc(), "_unroll_upsamplingDMA[{0},{1}]", srcOffset, nextOffset);
        VPURT::wrapIntoTaskOp<VPUIP::UpsamplingDMAOp>(
                rewriter, vpurtTask.getWaitBarriers(), vpurtTask.getUpdateBarriers(), newLoc, newSrcBuff, newDstBuff,
                upsamplingDMAOp.getUpsamplingFactorAttr(), descriptorAttr, upsamplingDMAOp.getExpandAttr(), dmaPort,
                upsamplingDMAOp.getIsOutOfOrder(), upsamplingDMAOp.getIsCritical(), upsamplingDMAOp.getDmaHwpIdAttr(),
                upsamplingDMAOp.getProfilingMetadataAttr());
        dmaPort = (dmaPort + 1) % _dmaPortCount;
        srcOffset = nextOffset;
        dstOffset += outShape.totalSize() * elemTypeSize.count();
    }
    rewriter.eraseOp(vpurtTask);
    return mlir::success();
}

//
// UnrollUpsamplingDMAPass
//

class UnrollUpsamplingDMAPass final : public VPUIP::UnrollUpsamplingDMABase<UnrollUpsamplingDMAPass> {
public:
    explicit UnrollUpsamplingDMAPass(Logger log) {
        Base::initLogger(log, Base::getArgumentName());
    }

private:
    void safeRunOnFunc() final;
};

void UnrollUpsamplingDMAPass::safeRunOnFunc() {
    auto& ctx = getContext();
    auto func = getOperation();
    auto module = func->getParentOfType<mlir::ModuleOp>();
    auto dmaOp = IE::getAvailableExecutor(module, VPU::ExecutorKind::DMA_NN);
    auto dmaPortCount = dmaOp.getCount();

    mlir::RewritePatternSet patterns(&ctx);
    patterns.insert<UpsamplingDMARewriter>(&ctx, dmaPortCount, _log);

    if (mlir::failed(
                mlir::applyPatternsAndFoldGreedily(func, std::move(patterns), vpux::getDefaultGreedyRewriteConfig()))) {
        signalPassFailure();
    }
}

}  // namespace

//
// createUnrollUpsamplingDMAPass
//

std::unique_ptr<mlir::Pass> vpux::VPUIP::createUnrollUpsamplingDMAPass(Logger log) {
    return std::make_unique<UnrollUpsamplingDMAPass>(log);
}
